#include "stm8s.h"
#include "eeprom.h"

// Records the current EEPROM address absolutely within a VGM file.
u16 EEPROM_CurrentAddressAbsolute;
// Records the current EEPROM address relative to the start of the current chip.
u16 EEPROM_CurrentAddressRelative;

// Stores bytes in a buffer for quicker retrieval in future.
u8 EEPROM_Buffer[EEPROM_BUFFER_SIZE];

// The EEPROM buffer "read" index.
u8 EEPROM_BufferRead;
// The EEPROM buffer "write" index.
u8 EEPROM_BufferWrite;
// The EEPROM number of bytes in the buffer.
u8 EEPROM_BufferCount;

// Initialises the IC periperal for reading from the EEPROM.
void EEPROM_Init(void) {
	I2C_DeInit();
	I2C_Cmd(ENABLE);
	I2C_Init(100000, 0xB0, I2C_DUTYCYCLE_2, I2C_ACK_CURR, I2C_ADDMODE_7BIT, 16);
}

// Start reading from the IC EEPROM at a particular address.
void EEPROM_StartReading(u16 address) {

	// Calculate the slave address of the IC EEPROM.
	u8 slaveAddress = EEPROM_BASE_ADDRESS + ((address / EEPROM_SIZE) << 1);

	// Store this for later.
	EEPROM_CurrentAddressAbsolute = address;	

	// Wrap the address around to fit within the size of one EEPROM.
	EEPROM_CurrentAddressRelative = address % EEPROM_SIZE;
	
	// Wait for the bus to become available.
	while(I2C_GetFlagStatus(I2C_FLAG_BUSBUSY));
	// Issue a START condition.
	I2C_GenerateSTART(ENABLE);
	while (!I2C_CheckEvent(I2C_EVENT_MASTER_START_SENT));
	// Transmit the EEPROM base address to write the read address.
	I2C_Send7bitAddress(slaveAddress, I2C_DIRECTION_TX);
	while (!I2C_CheckEvent(I2C_EVENT_MASTER_ADDRESS_ACKED));
	I2C_ClearFlag(I2C_FLAG_ADDRESSSENTMATCHED);
	// Send the address to start reading from.
	I2C_SendData((u8)(EEPROM_CurrentAddressRelative >> 8));
	while (!I2C_CheckEvent(I2C_EVENT_MASTER_BYTE_TRANSMITTED));
	I2C_SendData((u8)EEPROM_CurrentAddressRelative);
	while (!I2C_CheckEvent(I2C_EVENT_MASTER_BYTE_TRANSMITTED));
	
	// Generate another START condition to start reading.
	I2C_GenerateSTART(ENABLE);
	while (!I2C_CheckEvent(I2C_EVENT_MASTER_START_SENT));
	// Transmit the EEPROM base address to start reading.
	I2C_Send7bitAddress(slaveAddress, I2C_DIRECTION_RX);
	while (!I2C_CheckEvent(I2C_EVENT_MASTER_ADDRESS_ACKED));
	I2C_ClearFlag(I2C_FLAG_ADDRESSSENTMATCHED);
	
	// Flush the buffer.
	EEPROM_BufferRead = EEPROM_BufferWrite = EEPROM_BufferCount = 0;
}

// Read from the IC EEPROM at a particular address.
u8 EEPROM_ReadByte(void) {
	// Read the data.
	u8 data;
	while (!I2C_CheckEvent(I2C_EVENT_MASTER_BYTE_RECEIVED));
	data = I2C_ReceiveData();
	// Have we reached the end of the EEPROM?
	++EEPROM_CurrentAddressAbsolute;
	if (++EEPROM_CurrentAddressRelative >= EEPROM_SIZE) {
		// Yes, so re-seek.
		EEPROM_Seek(EEPROM_CurrentAddressAbsolute);
	}
	return data;
}

// Stop reading from the IC EEPROM.
void EEPROM_StopReading(void) {
	// Read a dummy byte without acknowledging.
	I2C_AcknowledgeConfig(I2C_ACK_NONE);
	while (!I2C_CheckEvent(I2C_EVENT_MASTER_BYTE_RECEIVED));
	I2C_ReceiveData();
	// Generate a stop condition.
	I2C_GenerateSTOP(ENABLE);
	// Wait for the bus to become available.
	while(I2C_GetFlagStatus(I2C_FLAG_BUSBUSY));
	// Re-enable the acknowledgement.
	I2C_AcknowledgeConfig(I2C_ACK_CURR);
}

// Seek to a new address in the IC EEPROM.
void EEPROM_Seek(u16 address) {
	EEPROM_StopReading();
	EEPROM_StartReading(address);
}

// Reads a byte from the buffer if available, directly from the EEPROM if not.
u8 EEPROM_ReadByteBuffered(void) {
	u8 data;
	if (EEPROM_BufferCount) {
		// Fetch a byte from the buffer.
		data = EEPROM_Buffer[EEPROM_BufferRead++];
		// Have we reached the end of the circular buffer?
		if (EEPROM_BufferRead == EEPROM_BUFFER_SIZE) {
			EEPROM_BufferRead = 0;
		}
		// One less byte in the buffer.
		--EEPROM_BufferCount;
	} else {
		// Read a byte directly.
		data = EEPROM_ReadByte();
	}
	return data;
}

// Read a byte from the EEPROM directly and top up the buffer with it if the buffer is not already full.
void EEPROM_TopUpBuffer(void) {
	// Only top up the buffer if it isn't already full.
	if (EEPROM_BufferCount < EEPROM_BUFFER_SIZE) {
		// Fetch a byte directly
		EEPROM_Buffer[EEPROM_BufferWrite++] = EEPROM_ReadByte();
		// Have we reached the end of the circular buffer?
		if (EEPROM_BufferWrite == EEPROM_BUFFER_SIZE) {
			EEPROM_BufferWrite = 0;
		}
		// One more byte in the buffer.
		++EEPROM_BufferCount;
	}
}

// Read a byte directly from a particular address.
u8 EEPROM_ReadByteAt(u16 address) {
	EEPROM_Seek(address);
	return EEPROM_ReadByte();
}

// Read an unsigned short (16 bits) directly from a particular address.
u16 EEPROM_ReadUShortAt(u16 address) {
	u16 result = EEPROM_ReadByteAt(address);
	result |= (EEPROM_ReadByte() << 8);
	return result;
}

// Read an unsigned integer (32 bits) directly from a particular address.
u32 EEPROM_ReadUIntAt(u16 address) {
	u32 result = EEPROM_ReadUShortAt(address);
	result |= ((u32)EEPROM_ReadByte() << (u32)16);
	result |= ((u32)EEPROM_ReadByte() << (u32)24);
	return result;
}