#include "STM8S105C_S.inc"

; ============================================================================ ;
; Variables                                                                    ;
; ============================================================================ ;
switch .ubsct

; PSG channel registers.
xref _PSG_Channels
_PSG_Channels:    ds.b 5*4

define_channel:macro
PSG_C\1Period:     equ _PSG_Channels+(5*\1+0)
PSG_C\1Amplitude:  equ _PSG_Channels+(5*\1+2)
PSG_C\1Counter:    equ _PSG_Channels+(5*\1+3)
endm

define_channel 0
define_channel 1
define_channel 2
define_channel 3

xdef _PSG_Output, _PSG_StereoMask
_PSG_Output:        ds.b 1 ; Stores the output level of each channel.
_PSG_StereoMask:    ds.b 1 

xdef _PSG_SampleTimer
_PSG_SampleTimer:   ds.w 1 ; Decrements every sample until it reaches 0.

xdef _PSG_ShiftRegister, _PSG_TappedBits, _PSG_ShiftRegisterWidth
_PSG_ShiftRegister: ds.w 1 ; Linear feedback shift register.
_PSG_TappedBits:    ds.w 1 ; Linear feedback shift register tapped bits.
_PSG_ShiftRegisterWidth: ds.b 1 ; Width of the shift register in bits.

ShiftCalculateEorA: ds.w 1 ; Used to calculate the parity of the linear feedback shift register.
ShiftCalculateEorB: ds.w 1 ; Used to calculate the parity of the linear feedback shift register.

SampleL:            ds.b 1 ; Stores the sample being generated for the left channel.
SampleR:            ds.b 1 ; Stores the sample being generated for the right channel.

ChannelAmplitude:   ds.b 1 ; Used to store the temporary amplitude for the current channel.

ClockStep:          ds.w 1 ; How many clock ticks to add to the tone registers each sample.

xdef _PSG_PendingCycles, _PSG_ClockRate
_PSG_PendingCycles: ds.w 1 ; Used to count the number of clock cycles we need to run the PSG emulator for.
_PSG_ClockRate:     ds.w 1 ; Clock speed of the PSG divided by 256.

; ============================================================================ ;
; Code                                                                         ;
; ============================================================================ ;

switch .text

; Updates a single tone register.
update_tone_counter:macro
	; Decrement the tone counter.
	ldw x,PSG_C\1Counter
	addw x,ClockStep
	; Has it overflowed?
	jrpl tone_\1_not_overflowed
	; It has overflowed!
	addw x,PSG_C\1Period
	bcpl _PSG_Output,#\1
tone_\1_not_overflowed:
	ldw PSG_C\1Counter,x
	; Add the level to the output if required.
	ld a,#15
	btjf _PSG_Output,#\1,tone_\1_not_active
	ld a,PSG_C\1Amplitude
tone_\1_not_active:
	ld ChannelAmplitude,a
	
	; Is the channel muted in the left?
	btjt _PSG_StereoMask,#(4+\1),tone_\1_not_muted_left
	ld a,#15
tone_\1_not_muted_left:
	add a,SampleL
	ld SampleL,a	
	
	ld a,ChannelAmplitude
	
	; Is the channel muted in the right?
	btjt _PSG_StereoMask,#(0+\1),tone_\1_not_muted_right
	ld a,#15
tone_\1_not_muted_right:
	add a,SampleR
	ld SampleR,a
endm

xor_shifted_self:macro
	ldw x,ShiftCalculateEorA
	if (\1)>3
	srlw x
	endif
	if (\1)>2
	srlw x
	endif
	if (\1)>1
	srlw x
	endif
	if (\1)>0
	srlw x
	endif
	ldw ShiftCalculateEorB,x
	ld a,ShiftCalculateEorA+0
	xor a,ShiftCalculateEorB+0
	ld ShiftCalculateEorA+0,a
	ld a,ShiftCalculateEorA+1
	xor a,ShiftCalculateEorB+1
	ld ShiftCalculateEorA+1,a
endm

xdef f_TIM3_UPD_OVF_IRQHandler
f_TIM3_UPD_OVF_IRQHandler:
	; Output the sample generated in the previous interrupt.
	ld a,SampleR
	ld TIM2_CCR1L,a
	ld a,SampleL
	ld TIM2_CCR2L,a
	
	; Reset the sample values.
	clr SampleL
	clr SampleR
	
	; Update the three tone channel counters.
	update_tone_counter 0
	update_tone_counter 1
	update_tone_counter 2
	
	; Update the noise channel.
	ldw x,PSG_C3Counter
	addw x,ClockStep
	ldw PSG_C3Counter,x
	jrpl noise_not_overflowed
	
	; Update the counter with its period.
	ld a,PSG_C3Period+1
	and a,#3
	jreq noise_0x10
	dec a
	jreq noise_0x20
	dec a
	jreq noise_0x40
noise_c2:
	addw x,PSG_C2Period
	jp noise_updated_counter
noise_0x40:
	addw x,#0x20 ;+0x10+0x10 = 0x40
noise_0x20:
	addw x,#0x10 ;+0x10 = 0x20
noise_0x10:
	addw x,#0x10
noise_updated_counter:
	ldw PSG_C3Counter,x

	; Noise channel operates at half the frequency of the others, so check its update flag.
	bcpl _PSG_Output,#4
	btjf _PSG_Output,#4,noise_not_overflowed
	
	; Are we generating periodic noise or white noise?
	btjt PSG_C3Period+1,#2,noise_white
noise_periodic:
	; Generating nice, easy periodic noise.
	ldw x,_PSG_ShiftRegister
	
	ld a,_PSG_ShiftRegisterWidth
	cp a,#16
	jreq noise_periodic_16

noise_periodic_15:
	srlw x
	ldw _PSG_ShiftRegister,x
	bccm _PSG_ShiftRegister+0,#6
	jp noise_output

noise_periodic_16:
	srlw x
	ldw _PSG_ShiftRegister,x
	bccm _PSG_ShiftRegister+0,#7
	jp noise_output
	
	
noise_white:
	; Generating not-quite-so-easy white noise.
	ldw x,_PSG_ShiftRegister
	ldw ShiftCalculateEorA,x
	
	; Tapped bits.
	ld a,ShiftCalculateEorA+0
	and a,_PSG_TappedBits+0
	ld ShiftCalculateEorA+0,a
	
	ld a,ShiftCalculateEorA+1
	and a,_PSG_TappedBits+1
	ld ShiftCalculateEorA+1,a
	
	; val ^= val >> 8
	ld a,ShiftCalculateEorA+1
	xor a,ShiftCalculateEorA+0
	ld ShiftCalculateEorA+1,a
	
	; val ^= val >> 4
	xor_shifted_self 4
	
	; val ^= val >> 2
	xor_shifted_self 2
	
	; val ^= val >> 1
	xor_shifted_self 1
	
	; Shift the shift register.
	ldw x,_PSG_ShiftRegister
	srlw x
	ldw _PSG_ShiftRegister,x
	
	; Top up the shift register.
	ld a,_PSG_ShiftRegisterWidth
	cp a,#16
	jreq noise_top_up_16
noise_top_up_15:
	srl ShiftCalculateEorA+1
	bccm _PSG_ShiftRegister+0,#6
	jp noise_output
noise_top_up_16:
	srl ShiftCalculateEorA+1
	bccm _PSG_ShiftRegister+0,#7
noise_output:
	bres _PSG_Output,#3
	btjf _PSG_ShiftRegister,#0,noise_output_off
	bset _PSG_Output,#3
noise_output_off:


noise_not_overflowed:
	; Generate the noise channel sample level.
	ld a,#15
	btjf _PSG_Output,#3,noise_output_not_active
	ld a,PSG_C3Amplitude
noise_output_not_active:
	ld ChannelAmplitude,a
	
	; Is the channel muted on the left?
	btjt _PSG_StereoMask,#(4+3),noise_not_muted_left
	ld a,#15
noise_not_muted_left:
	add a,SampleL
	ld SampleL,a
	
	ld a,ChannelAmplitude
	
	; Is the channel muted on the right?
	btjt _PSG_StereoMask,#(0+3),noise_not_muted_right
	ld a,#15
noise_not_muted_right:
	add a,SampleR
	ld SampleR,a

	; Decrement the sample timer.
	ldw x,_PSG_SampleTimer
	jreq sample_timer_depleted
	decw x
	ldw _PSG_SampleTimer,x
sample_timer_depleted:

	; Calculate the ClockStep
	ld a,#-1
	ld ClockStep+0,a
	clr a
	
	ldw x,_PSG_PendingCycles
	subw x,_PSG_ClockRate
	
calculate_clock_step_loop:
	dec a
	addw x,#2762
	jrmi calculate_clock_step_loop
	
	ldw _PSG_PendingCycles,x
	ld ClockStep+1,a

	; Clear interrupt pending flag.
	bres TIM3_SR1, #0
	iret